from typing import Dict, Set, Iterable, List
import pandas as pd
from doenut.data.modifiers.data_set_modifier import DataSetModifier


class DuplicateRemover(DataSetModifier):
    """
    Parses a dataset and removes all but the _first_ instance of any row that
    has duplicate values for the _inputs_. Will also remove the corresponding
    row in the responses.
    """

    @classmethod
    def _get_duplicate_rows(cls, data: pd.DataFrame) -> Dict[int, Set[int]]:
        duplicates = [x for x in data[data.duplicated()].index]
        results = {}
        for duplicate in duplicates:
            # find first row it is a dupe of
            for index in data.index:
                duplicate_row = data.iloc[duplicate]
                if index >= duplicate:
                    # pandas ensures the duplicate is later in the
                    # dataframe than the row it is a duplicate of.
                    raise OverflowError(
                        f"Duplicate is missing for {duplicate}"
                    )
                if data.iloc[index].equals(duplicate_row):
                    try:
                        results[index].add(duplicate)
                    except KeyError:
                        results[index] = {duplicate}
                    break
        return results

    @classmethod
    def _get_non_duplicate_rows(
        cls,
        data: pd.DataFrame,
        duplicates_dict: Dict[int, Iterable[int]] = None,
    ) -> List[int]:
        if duplicates_dict is None:
            # assume we are removing according to this dataset
            duplicates_dict = cls._get_duplicate_rows(data)

        # build the list of rows we don't want.
        duplicate_indices = set()
        for duplicate_set in duplicates_dict.values():
            duplicate_indices = duplicate_indices.union(duplicate_set)
        non_duplicates = [x for x in data.index if x not in duplicate_indices]
        return non_duplicates

    def __init__(self, inputs: pd.DataFrame, responses: pd.DataFrame) -> None:
        """
        This modifier will remove all rows from the dataset which have
        identical values for the _inputs_. The first instance in the dataset
        of a given set of values will be the one retained.
        @param inputs: The inputs of the dataset
        @param responses: The responses of the dataset
        """
        super().__init__(inputs, responses)
        # use input data to determine which rows are duplicates

        self.duplicate_dict = self._get_duplicate_rows(inputs)
        self.non_duplicate_rows = self._get_non_duplicate_rows(
            inputs, self.duplicate_dict
        )

    def apply_to_inputs(self, data: pd.DataFrame) -> pd.DataFrame:
        return data.iloc[self.non_duplicate_rows]

    def apply_to_responses(self, data: pd.DataFrame) -> pd.DataFrame:
        return data.iloc[self.non_duplicate_rows]



from abc import abstractmethod, ABC
import pandas as pd

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from doenut.data.modifiable_data_set import ModifiableDataSet


class DataSetModifier(ABC):
    """
    Parent class for all types of modifier.
    They take a dataset in, perform some form of operation on it and then
    pass it along
    """

    def __init__(
        self, inputs: pd.DataFrame, responses: pd.DataFrame, **kwargs
    ):
        """
        Does nothing, but defines the constructor for all other DataSets
        @param inputs: the processed inputs up till this point
        @param responses: the processed responses up till this point
        Use this to do things like check the size and ranges of the dataset.
        @param kwargs: any other arguments the modifier needs
        """
        pass

    @abstractmethod
    def apply_to_inputs(self, data: pd.DataFrame) -> pd.DataFrame:
        """
        Applies the modifier to the inputs of the dataset.
        @param data: The input data
        @return: The data post modification.
        """
        pass

    @abstractmethod
    def apply_to_responses(self, data: pd.DataFrame) -> pd.DataFrame:
        """
        Applies the modifier to the responses of the dataset.
        @param data: The response data
        @return: The data post modification.
        """
        pass



"""
DoENUT: data.modifiers
These classes provide ways to manipulate dataset (filtering, scaling, etc).
"""
from .ortho_scaler import OrthoScaler
from .column_selector import ColumnSelector
from .duplicate_remover import DuplicateRemover
from .duplicate_averager import DuplicateAverager



import pandas as pd


class DataSet:
    """
    A dataset that has had all it's modifiers applied.
    """

    def __init__(self, inputs, responses):
        self.inputs = inputs
        self.responses = responses

    def get_inputs(self) -> pd.DataFrame:
        return self.inputs

    def get_responses(self) -> pd.DataFrame:
        return self.responses



from typing import Dict, Iterable, List
import pandas as pd
from doenut.data.modifiers.duplicate_remover import DuplicateRemover


class DuplicateAverager(DuplicateRemover):
    """
    Parses a dataset and removes all but the _first_ instance of any row that
    has duplicate values for the _inputs_. Will also remove the corresponding
    row in the responses, replacing the remaining response with the averages
    of the duplicates' values.
    """

    @classmethod
    def _apply(
        cls,
        data: pd.DataFrame,
        duplicate_dict: Dict[int, Iterable[int]],
        non_duplicate_rows: List[int],
    ) -> pd.DataFrame:
        # first build a copy of the data with the duplicates removed
        results = data.iloc[non_duplicate_rows].copy(True)
        # now figure out the averages for the ones that need averaging
        for idx, dupes in duplicate_dict.items():
            to_average = [data.iloc[dupe] for dupe in dupes]
            to_average.append(results.iloc[idx])
            results.iloc[idx] = pd.concat(to_average, axis=1).T.mean(axis=0)
        return results

    def __init__(self, inputs: pd.DataFrame, responses: pd.DataFrame) -> None:
        """
        This modifier will remove all rows from the dataset which have
        identical values for the _inputs_, and set the response value to be
         the average of all the duplicates. The first instance in the dataset
        of a given set of values will be the one retained.
        @param inputs: The inputs of the dataset
        @param responses: The responses of the dataset
        """
        super().__init__(inputs, responses)

    def apply_to_inputs(self, data: pd.DataFrame) -> pd.DataFrame:
        return self._apply(data, self.duplicate_dict, self.non_duplicate_rows)

    def apply_to_responses(self, data: pd.DataFrame) -> pd.DataFrame:
        return self._apply(data, self.duplicate_dict, self.non_duplicate_rows)



import pandas as pd
from sklearn.linear_model import LinearRegression
from doenut.data.data_set import DataSet


class Model:
    """
    A simple linear regression model.

    This mostly exists as a base - you probably want AveragedModel
    """

    def __init__(self, data: DataSet, fit_intercept: bool) -> None:
        """
        Generate a simple model from the given values
        @param data: The inputs and responses to the model
        @param fit_intercept: Whether to fit the intercept to zero
        """
        self.data = data
        inputs = self.data.get_inputs()
        responses = self.data.get_responses()
        self.model = LinearRegression(fit_intercept=fit_intercept)
        self.model.fit(inputs, responses)
        self.predictions = self.get_predictions_for(inputs)
        self.r2 = self.get_r2_for(self.data)

    def get_predictions_for(self, inputs: pd.DataFrame) -> pd.DataFrame:
        """
        Generates the predictions of the model for a set of inputs
        @param inputs: The inputs to test against
        @return: the predictions from the model
        """
        return self.model.predict(inputs)

    def get_r2_for(self, data: DataSet):
        """
        Calculate the R2 Pearson coefficient for a given pairing of
        inputs and responses.
        @param data: The data to test.
        @return: the calculated R2 value as a float
        """
        return self.model.score(data.get_inputs(), data.get_responses())



from typing import List, Any

from doenut.data import ModifiableDataSet
from doenut.models import Model


class ModelSet:
    """
    Class to train and hold a group of related models.
    """

    def __init__(
        self,
        default_inputs=None,
        default_responses=None,
        default_scale_data=True,
        default_fit_intercept=True,
    ):
        self.default_inputs = default_inputs
        self.default_responses = default_responses
        self.default_scale_data = default_scale_data
        self.default_fit_intercept = default_fit_intercept
        self.models = []

    def _validate_value(self, name: str, value: Any = None) -> Any:
        if value is not None:
            return value
        default_name = f"default_{name}"
        if hasattr(self, default_name):
            value = getattr(self, default_name)
            if value is not None:
                return value
        raise ValueError(f"model set lacks default value for {name}")

    def add_model(
        self, inputs=None, responses=None, scale_data=None, fit_intercept=None
    ):
        inputs = self._validate_value("inputs", inputs)
        responses = self._validate_value("responses", responses)
        scale_data = self._validate_value("scale_data", scale_data)
        fit_intercept = self._validate_value("fit_intercept", fit_intercept)
        dataset = ModifiableDataSet(inputs, responses).get()
        model = Model(dataset, fit_intercept)
        self.models.append(model)
        return model

    def get_r2s(self):
        # return [x.r2 for x in self.models]
        return self.get_attributes("r2")

    def get_attributes(self, attribute: str) -> List:
        """
        Get a specified attribute from each model.
        Frustratingly, some are in the model, others in the model object.
        @param attribute:
        @return:
        """
        if hasattr(self.models[0], attribute):
            return [getattr(x, attribute) for x in self.models]
        if hasattr(self.models[0].model, attribute):
            return [getattr(x.model, attribute) for x in self.models]
        raise ValueError(f"Attribute {attribute} is not in the models")



from typing import Tuple

import numpy as np
import pandas as pd
import doenut
import copy

from doenut.data.modifiable_data_set import ModifiableDataSet
from doenut.models.model_set import ModelSet
from doenut.models.model import Model


class AveragedModel(Model):
    """
    Model scored as the average of multiple models generated from a single
    set of inputs via a leave-one-out approach.
    """

    @classmethod
    def tune_model(
        cls,
        data: ModifiableDataSet,
        fit_intercept: bool = True,
        response_key: str = None,
        drop_duplicates: str = "yes",
    ) -> Tuple["AveragedModel", "AveragedModel"]:
        """
        Generate a pair of models from the same set of data. One using scaled
        data the other unscaled.

        The scaled model can then be used for determining which columns to drop
        for later models, and the unscaled model for checking the models
        performance against validation data (or just for using once done).

        @param data: The dataset to test against. This should be unscaled.
        @param fit_intercept: Whether to fit the intercept or not (usually yes)
        @param response_key: If there are more than one response columns,
        which to use.
        @param drop_duplicates: Whether to apply any duplicate reduction to
        the data as it is trained. Should be one of 'yes', 'no' or 'average'
        @return: A tuple of AveragedModels: (scaled, unscaled)
        """
        scaled_model = AveragedModel(
            data,
            scale_data=True,
            scale_run_data=True,
            fit_intercept=fit_intercept,
            response_key=response_key,
            drop_duplicates=drop_duplicates,
        )
        unscaled_model = AveragedModel(
            data,
            scale_data=False,
            scale_run_data=False,
            fit_intercept=fit_intercept,
            response_key=response_key,
            drop_duplicates=drop_duplicates,
        )
        return scaled_model, unscaled_model

    def __init__(
        self,
        data: ModifiableDataSet,
        scale_data: bool = True,
        scale_run_data: bool = True,
        fit_intercept: bool = True,
        response_key: str = None,
        drop_duplicates: str = "yes",
    ):
        """
        Constructor
        @param data: the data to run / test against.
        @param scale_data: Whether to scale the overall data before running it.
        @param scale_run_data: Whether to normalise the data for each run
        @param fit_intercept: Whether to fit the intercept to zero
        @param response_key: for multi-column responses, which one to test on
        @param drop_duplicates: whether to drop duplicate values or not.
        May also be 'average' which will cause them to be dropped, but the one
        left will have its response value(s) set to the average of all the
        duplicates.
        """
        proc_data = copy.deepcopy(data)
        if scale_data:
            proc_data.scale(False)
        # Call super to set up basic model
        super().__init__(proc_data.get(), fit_intercept)

        # check the columns
        responses = self.data.get_responses()
        if response_key is None:
            if len(responses.columns) > 1:
                raise ValueError(
                    "No response key specified and multiple response columns"
                )
            response_key = responses.columns[0]

        # Get the processed inputs + responses (after filtering + dedupe
        proc_inputs, proc_responses = None, None
        if isinstance(drop_duplicates, str):
            if str.lower(drop_duplicates) == "yes":
                proc_data.drop_duplicates()
            elif str.lower(drop_duplicates) == "average":
                proc_data.average_duplicates()
            elif str.lower(drop_duplicates) == "no":
                pass
            else:
                raise ValueError(
                    f"Invalid drop_duplicates value {drop_duplicates}"
                    " - should one of 'yes', 'no', 'average'"
                )

        final_data = proc_data.get()
        proc_inputs = final_data.get_inputs()
        proc_responses = final_data.get_responses()

        # Use leave-one-out on the input data rows to generate a set of models
        self.models = ModelSet(None, None, fit_intercept)
        model_predictions = []
        errors = []
        model_responses = []
        for i, row_idx in enumerate(proc_inputs.index):
            test_input = proc_inputs.iloc[i].to_numpy().reshape(1, -1)
            test_response = proc_responses.iloc[i]
            train_input = proc_inputs.drop(row_idx).to_numpy()
            train_responses = proc_responses.drop(row_idx)
            # We need to re-scale each column, using the training data *only*,
            # but then applying the same scaling to the test data.
            if scale_run_data:
                train_input, mj, rj = doenut.orthogonal_scaling(train_input, 0)
                test_input = doenut.scale_by(test_input, mj, rj)
            model = self.models.add_model(
                train_input, train_responses, False, fit_intercept
            )
            predictions = model.get_predictions_for(test_input)[0]
            model_predictions.append(predictions)
            model_responses.append(test_response)
            errors.append(test_response - predictions)
        self.coeffs = self.models.get_attributes("coef_")
        self.intercepts = self.models.get_attributes("intercept_")
        self.averaged_coeffs = np.mean(np.array(self.coeffs), axis=0)
        self.averaged_intercepts = np.mean(np.array(self.intercepts), axis=0)
        self.r2s = self.models.get_r2s()

        # replace our initial model with the averaged one to determine R2/Q2.
        self.model.coef_ = self.averaged_coeffs
        self.model.intercept_ = self.averaged_intercepts

        self.r2 = self.get_r2_for(final_data)

        # Now calculate q2
        self.q2_predictions = pd.DataFrame.from_records(
            model_predictions, columns=proc_responses.columns
        )
        self.q2_ground_truths = pd.DataFrame.from_records(
            model_responses, columns=proc_responses.columns
        )
        self.q2 = doenut.Calculate_Q2(
            self.q2_ground_truths,
            self.q2_predictions,
            proc_responses,
            response_key,
        )
        # finally make a fitted model.
        start_data = data.get()
        self.model.fit(start_data.get_inputs(), start_data.get_responses())
        self.predictions = self.get_predictions_for(proc_inputs)



"""
DoENUT

Design of Experiments Numerical Utility Toolkit

DoENUT is a set of classes and functions designed to make Design of Experiments
easier in Python.

To get started, see the workbooks under Tutorials or look at AveragedModel
and ModifiableDataSet.

As a very quick start, assuming your data is split into a pair of pandas
DataFrames, one for the input data and one for the responses, the following
code will create a standard model and generate some stats on how good it is.

dataset = doenut.data.ModifiableDataSet(inputs, responses)
model = doenut.models.AveragedModel(dataset)
r2, q2 = model.r2, model.q2
print(f"R2 is {r2}, Q2 is {q2}")
doenut.plot.plot_summary_of_fit_small(r2, q2)
doenut.plot.coeff_plot(model.coeffs,
                       labels=list(dataset.get().inputs.columns),
                       errors='p95',
                       normalise=True)
"""

from importlib.metadata import version

__version__ = version("doenut")

from . import doenut
from . import designer
from . import plot
from . import models

from .doenut import (
    orthogonal_scaling,
    add_higher_order_terms,
    autotune_model,
    scale_by,
    Calculate_Q2,
    calc_ave_coeffs_and_errors,
    map_chemical_space,
    predict_from_model,
)



"""
DoENUT: models
A model is an approximation of a system, built by fitting various equations to
a dataset.

Currently, this module provides four classes - two model classes and two model
grouping classes.

Model is a basic fitted via linear regression model. It's a fairly thin layer
over sklearn's model and the base for the rest of the code.

AveragedModel is the more useful one. As well as generating a normal model, it
will also generate a set of fitting models via a leave-one-out methodology.
This allows it to calculate the Q2 cross validation correlation coefficient.

ModelSet can be used when you have multiple columns/parameters in your response
data. It will generate one Model for each response column. Similarly,
AveragedModelSet will generate one AveragedModel for each response column.

"""
from .model import Model
from .averaged_model import AveragedModel
from .model_set import ModelSet
from .averaged_model_set import AveragedModelSet



import pandas as pd

from doenut.data import ModifiableDataSet
from doenut.models import ModelSet
from doenut.models import AveragedModel


class AveragedModelSet(ModelSet):
    @classmethod
    def multiple_response_columns(
        cls,
        inputs: pd.DataFrame = None,
        responses: pd.DataFrame = None,
        scale_data: bool = True,
        scale_run_data: bool = True,
        fit_intercept: bool = True,
        drop_duplicates: str = "yes",
        input_selector: list = [],
    ) -> "AveragedModelSet":
        result = AveragedModelSet(
            inputs,
            responses,
            scale_data,
            scale_run_data,
            fit_intercept,
            [],
            drop_duplicates,
            input_selector,
        )
        for column in responses.columns:
            result.add_model(response_key=column)
        return result

    def __init__(
        self,
        default_inputs: pd.DataFrame = None,
        default_responses: pd.DataFrame = None,
        default_scale_data: bool = True,
        default_scale_run_data: bool = True,
        default_fit_intercept: bool = True,
        default_response_key: list = [0],
        default_drop_duplicates: str = "yes",
        default_input_selector: list = [],
    ):
        super().__init__(
            default_inputs,
            default_responses,
            default_scale_data,
            default_fit_intercept,
        )
        self.default_scale_run_data = default_scale_run_data
        self.default_response_key = default_response_key
        self.default_drop_duplicates = default_drop_duplicates
        self.default_input_selector = default_input_selector

    def add_model(
        self,
        inputs=None,
        responses=None,
        scale_data=None,
        scale_run_data=None,
        fit_intercept=None,
        response_key=None,
        drop_duplicates=None,
        input_selector=None,
    ):
        inputs = self._validate_value("inputs", inputs)
        responses = self._validate_value("responses", responses)
        scale_data = self._validate_value("scale_data", scale_data)
        scale_run_data = self._validate_value("scale_run_data", scale_run_data)
        fit_intercept = self._validate_value("fit_intercept", fit_intercept)
        response_key = self._validate_value("response_key", response_key)
        drop_duplicates = self._validate_value(
            "drop_duplicates", drop_duplicates
        )
        input_selector = self._validate_value("input_selector", input_selector)

        data = ModifiableDataSet(inputs, responses)
        # if scale_data:
        #     data.scale()
        if input_selector:
            data.filter(input_selector)
        model = AveragedModel(
            data,
            scale_data,
            scale_run_data,
            fit_intercept,
            response_key,
            drop_duplicates,
        )
        self.models.append(model)
        return model



"""
DoENUT: data
These classes handle the grouping and storage of DataSets for DoE.

A DataSet consists of a set of inputs and a set of responses.
This module provides two classes - a basic DataSet and ModifiableDataSet.

Mostly, you want to use ModifiableDataSet as that allows for the operations
you will need for DoE.
"""
from .modifiable_data_set import ModifiableDataSet
from .data_set import DataSet



import copy
from typing import Type, List

import pandas as pd

from doenut.data.data_set import DataSet
from doenut.data.modifiers.column_selector import ColumnSelector
from doenut.data.modifiers.duplicate_averager import DuplicateAverager
from doenut.data.modifiers.duplicate_remover import DuplicateRemover
from doenut.data.modifiers.ortho_scaler import OrthoScaler
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from doenut.data.modifiers.data_set_modifier import DataSetModifier


class ModifiableDataSet:
    """
    Typically when doing DoE you will want to apply various modifiers such as
    scaling or filtering of columns to your dataset. ModifiableDataSet is
    DoENUT's mechanism to provide this.

    A base dataset consists of two pandas dataframes, one for the inputs and
    one for the responses. These should be of the same length.

    Once you have built the dataset you can add modifiers to it using
    add_modifier or (more likely) via the helper functions such as filter and
    scale. Finally, get() will then give you a DataSet object containing the
    result of applying all the modifiers.

    Be aware that modifiers are applied in the order they are added, and that
    modifiers cannot be removed once added. ModifiableDataSet makes deep
    copies of the dataframes, so the original data will not get changed.

    All the modifier functions return a link to self, so they can be used as
    per the builder pattern - i.e. so you can write code like:
      dataset = ModifiableDataset(inputs,responses).filter(list).scale()
    """

    def __init__(self, inputs: pd.DataFrame, responses: pd.DataFrame) -> None:
        if inputs is None or len(inputs) == 0:
            raise ValueError("Inputs must not be empty")
        if responses is None or len(responses) == 0:
            raise ValueError("Responses must not be empty")
        if len(inputs) != len(responses):
            raise ValueError("Inputs and Responses must have the same length")

        # the input values
        self.inputs = inputs
        self.responses = responses
        # the input values post (current) processing
        self._proc_inputs = copy.deepcopy(inputs)
        self._proc_responses = copy.deepcopy(responses)
        self.modifiers = []

    def get(self) -> DataSet:
        return DataSet(self._proc_inputs, self._proc_responses)

    def add_modifier(
        self, modifier: Type["DataSetModifier"], **kwargs
    ) -> "ModifiableDataSet":
        """
        Adds a new modifier to the stack.
        @param modifier: The new modifier to add
        @param kwargs: Any additional arguments the modifier is expecting.
        """
        modifier = modifier(self._proc_inputs, self._proc_responses, **kwargs)
        self.modifiers.append(modifier)
        self._proc_inputs = modifier.apply_to_inputs(self._proc_inputs)
        self._proc_responses = modifier.apply_to_responses(
            self._proc_responses
        )
        return self

    def filter(
        self,
        input_selector: List["str | int"] = None,
        response_selector: List["str | int"] = None,
    ) -> "ModifiableDataSet":
        """
        Select a subset of the columns in this dataset.
        You must specify at least one selector.
        Each select selector can be either a list of column names or indices
        that you wish to keep.
        @param input_selector: Filter for the input data
        @param response_selector: Filter for the response data
        @return: this dataset
        """
        return self.add_modifier(
            ColumnSelector,
            input_selector=input_selector,
            response_selector=response_selector,
        )

    def scale(self, scale_responses: bool = False) -> "ModifiableDataSet":
        """
        Apply an orthographic scaling to the dataset
        i.e. apply a linear scaling so each column is in the range -1...1
        @param scale_responses: Whether to scale the reponse data as well
        @return: this dataset
        """
        return self.add_modifier(OrthoScaler, scale_responses=scale_responses)

    def drop_duplicates(self) -> "ModifiableDataSet":
        """
        Removes all duplicate rows from the dataset. The first instance of
        each duplicate will be kept.
        NOTE: while only the inputs are considered for whether a row is a
        duplicate or now, duplicates will be removed from both inputs and
        responses.
        @return: self
        """
        return self.add_modifier(DuplicateRemover)

    def average_duplicates(self) -> "ModifiableDataSet":
        """
        Removes all duplicate rows from the dataset. The first instance of
        each duplicate will be kept, and it's responses set to the average of
        all the rows that matched it.
        NOTE: only inputs values are considered for whether a row is a
        duplicate or not
        @return: self
        """
        return self.add_modifier(DuplicateAverager)



############################################################################################################
#
#              DoENUT Designer
#
############################################################################################################
import doepy.build

# !!! TO-DO !!!
#
# make this into a nice proper class
import pandas as pd
import numpy as np
import copy


def _check_is_input_dict(data):
    """
    Most of these functions require a dictionary of lists as their input data
    This is a helper function that will throw an appropriate assert if needed.
    """
    if not isinstance(data, dict):
        raise TypeError("Input data must be a dictionary")
    for key, value in data.items():
        try:
            _ = iter(value)
        except TypeError as e:
            print(f"Parameter {key} is not iterable")
            raise e


def get_ranges(data):
    """
    Go through a dictionary of value lists, and return the same, but with
    only the min / max value from each in each.
    """
    # first check we are being passed something sane
    _check_is_input_dict(data)

    result = {}
    for key, value in data.items():
        result[key] = [min(value), max(value)]
    return result


def full_fact(data):
    """
    Generate a full factorial model from the supplied parameters
    :param data: dict of lists of allowed values for each parameter
    :return: a panda dataframe of the all the experiments
    """
    # first validate the inputs are all lists or list like
    # while we are here, work out how bit this is.
    row_count = 1
    for key, value in data.items():
        try:
            _ = iter(value)
        except TypeError as e:
            print(f"Parameter {key} is not iterable")
            raise e
        row_count = row_count * len(value)
    result = np.zeros((row_count, len(data.keys())), dtype="O")

    # Now build up the data column by column
    # how many row 'blocks' there are to the left of the current column
    left_data = 1
    # how many rows are remaining to fill.
    right_data = row_count

    # Note, there are a lot of int() calls below,
    # but these should always be valid as we are dividing ints by other ints
    # that are divisors of it.
    for column_idx, (column, values) in enumerate(data.items()):
        value_count = len(values)
        # how many times do we need to write this value in a row?
        rows_per_value = int(right_data / value_count)
        for group in range(left_data):
            # work out where this group begins
            offset = rows_per_value * value_count * group
            for idx, value in enumerate(values):
                start = int(offset + (idx * rows_per_value))
                end = int(start + rows_per_value)
                result[start:end, column_idx] = value
        # re-establish the invariants
        left_data = int(left_data * value_count)
        right_data = right_data / value_count

    result = pd.DataFrame(columns=list(data.keys()), dtype=object, data=result)
    return result


def frac_fact(data, resolution=None):
    """
    build a 2-level fractional factorial design
    :param data dictionary to design from
    :param resolution what resolution model to build. Default is param_count/2
    """
    _check_is_input_dict(data)
    if resolution is None:
        resolution = int(len(data.keys()) / 2) + 1
    if resolution >= len(data.keys()):
        raise ValueError(
            "Resolution has to be less than the number of parameters"
        )
    if resolution == 1:
        raise ValueError("Resolution of 1 is meaningless")

    # only want the limits
    data_ranges = get_ranges(data)

    # TODO:: Now implement the hard bit!
    return doepy.build.frac_fact_res(data_ranges, resolution)


# TODO:: this should be a base class with hte actual experiment overwritten in the sub-class
def experiment_designer(
    levels, res, do_midpoints=True, shuffle=True, repeats=1, num_midpoints=3
):
    """
    levels is a dictionary of factor name and levels
    res is the resolution (for frac fact) - shouldn't be in class
    do_midpoints whether to add in the mid points
    shuffle whether to shuffle
    repeats how many repeats you're doing of the NON-MIDPOINTS
    num_midpoints, how many midpoints to do
    """

    # deepcopy as their code overwrites the levels >:(
    levels_in = copy.deepcopy(levels)
    design = frac_fact_res(levels_in, res=res)
    factor_names = [x for x in levels.keys()]
    if repeats > 1:
        for i in range(repeats):
            design = design.append(midpoints, ignore_index=True)
    if do_midpoints:
        midpoints = {}
        for factor in levels.keys():
            if len(levels[factor]) > 2:
                midpoints[factor] = np.median(levels[factor])
            else:
                midpoints[factor] = np.mean(levels[factor])
        # midpoints = pd.DataFrame(midpoints, index=str(len(design)+1))
        for i in range(num_midpoints):
            design = design.append(midpoints, ignore_index=True)

        if shuffle:
            design = design.sample(frac=1)

    return design


def frac_fact_res_designer(
    levels, res, do_midpoints=True, shuffle=True, repeats=1, num_midpoints=3
):
    levels_in = copy.deepcopy(levels)
    design = frac_fact_res(levels_in, res=res)
    factor_names = [x for x in levels.keys()]
    if do_midpoints:
        midpoints = {}
        for factor in levels.keys():
            if len(levels[factor]) > 2:
                midpoints[factor] = np.median(levels[factor])
            else:
                midpoints[factor] = np.mean(levels[factor])
        # midpoints = pd.DataFrame(midpoints, index=str(len(design)+1))
        for i in range(num_midpoints):
            design = design.append(midpoints, ignore_index=True)

        if shuffle:
            design = design.sample(frac=1)

    return design


def fact_designer(
    levels, do_midpoints=True, shuffle=True, repeats=1, num_midpoints=3
):
    levels_in = copy.deepcopy(levels)
    # Build a basic full factorial design.
    design = full_fact(levels_in)
    if do_midpoints:
        midpoints = {}
        for factor in levels.keys():
            if len(levels[factor]) > 2:
                midpoints[factor] = np.repeat(
                    np.median(levels[factor]), num_midpoints
                )
            else:
                midpoints[factor] = np.repeat(
                    np.mean(levels[factor]), num_midpoints
                )
        midpoint_df = pd.DataFrame.from_dict(midpoints)
        design = pd.concat([design, midpoint_df], ignore_index=True)

        if shuffle:
            design = design.sample(frac=1)

    return design



"""
DOENUT
Design of Experiments Numerical Utility Toolkit
"""

# first we import some useful libraries
import numpy as np
import pandas as pd
import copy
from sklearn.linear_model import LinearRegression
from typing import Tuple
from doenut.data import ModifiableDataSet
from doenut.models import AveragedModel


def orthogonal_scaling(inputs, axis=0):
    # the scaling thingy that Modde uses
    inputs_max = np.max(inputs, axis)
    inputs_min = np.min(inputs, axis)
    Mj = (inputs_min + inputs_max) / 2
    Rj = (inputs_max - inputs_min) / 2
    scaled_inputs = (inputs - Mj) / Rj
    return scaled_inputs, Mj, Rj


def scale_1D_data(scaler, data, do_fit=True):
    if do_fit:
        scaler.fit(data.reshape(-1, 1))
    data_scaled = scaler.transform(data.reshape(-1, 1))
    data_scaled = data_scaled.reshape(-1)
    return data_scaled, scaler


def scale_by(new_data, Mj, Rj):
    # the scaling thingy that Modde uses
    # TODO:: Any form of sanity checking whatsoever.
    new_data = (new_data - Mj) / Rj
    return new_data


def find_replicates(inputs):
    """Find experimental settings that are replicates"""
    # list comps ftw!
    a = [x for x in inputs[inputs.duplicated()].index]
    b = [x for x in inputs[inputs.duplicated(keep="last")].index]
    replicate_row_list = np.unique(np.array(a + b))
    return replicate_row_list


def train_model(
    inputs,
    responses,
    test_responses,
    do_scaling_here=False,
    fit_intercept=False,
    verbose=True,
):
    """A simple function to train a model
    :param inputs: full set of terms for the model (x_n)
    :param responses: expected responses for the inputs (ground truth, y)
    :param test_responses: expected responses for seperate test data (if used)
    :param do_scaling_here: whether to scale the data
    :param fit_intercept: whether to fit the intercept
    :param verbose: whether to perform additional logging.
    :return: A tuple of:
        A model fitted to the data,
        the inputs used
        the R2 of that model
        the predictions that model makes for the original inputs
    """
    if do_scaling_here:
        inputs, _, _ = orthogonal_scaling(inputs, axis=0)
    if test_responses is None:
        test_responses = responses
    model = LinearRegression(fit_intercept=fit_intercept)
    model.fit(inputs, responses)
    predictions = model.predict(inputs)
    R2 = model.score(inputs, test_responses)
    if verbose:
        print("R squared for this model is {:.3}".format(R2))
    return model, inputs, R2, predictions


def Calculate_R2(ground_truth, predictions, key, word="test", verbose=True):
    """Calculates R2 from input data
    You can use this to calculate q2 if you're
    using the test ground truth as the mean
    else use calculate Q2
    I think this is what Modde uses for PLS fitting"""
    errors = ground_truth[[key]] - predictions[[key]]
    test_mean = np.mean(ground_truth[[key]], axis=0)
    if verbose:
        print(f"Mean of {word} set: {test_mean[0]}")
    errors["squared"] = errors[key] * errors[key]
    sum_squares_residuals = sum(errors["squared"])
    if verbose:
        print(
            "Sum of squares of the residuals (explained variance) is"
            f"{sum_squares_residuals}"
        )
    sum_squares_total = sum((ground_truth[key] - test_mean[0]) ** 2)
    if verbose:
        print(f"Sum of squares total (total variance) is {sum_squares_total}")
    R2 = 1 - (sum_squares_residuals / sum_squares_total)
    if word == "test":
        print("{} is {:.3}".format("Q2", R2))
    else:
        print("{} is {:.3}".format("R2", R2))
    return R2


def Calculate_Q2(
    ground_truth, predictions, train_responses, key, word="test", verbose=True
):
    """A different way of calculating Q2
    this uses the mean from the training data, not the
    test ground truth"""
    errors = ground_truth[[key]] - predictions[[key]]
    train_mean = np.mean(train_responses[[key]], axis=0)
    test_mean = np.mean(ground_truth[[key]], axis=0)
    if verbose:
        print(f"Mean of {word} set: {test_mean.iloc[0]}")
        print(f"Mean being used: {train_mean.iloc[0]}")
    errors["squared"] = errors[key] * errors[key]
    sum_squares_residuals = sum(errors["squared"])
    if verbose:
        print(
            "Sum of squares of the residuals (explained variance) is"
            f"{sum_squares_residuals}"
        )
    sum_squares_total = sum((ground_truth[key] - train_mean.iloc[0]) ** 2)
    # stuff from Modde
    # errors/1

    if verbose:
        print(f"Sum of squares total (total variance) is {sum_squares_total}")
    R2 = 1 - (sum_squares_residuals / sum_squares_total)
    if word == "test":
        print("{} is {:.3}".format("Q2", R2))
    else:
        print("{} is {:.3}".format("R2", R2))
    return R2


def dunk(setting=None):
    if setting == "coffee":
        print("Splash!")
    else:
        print("Requires coffee")
    return


def average_replicates(
    inputs: pd.DataFrame, responses: pd.DataFrame, verbose: bool = False
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """averages inputs that are the same
    TO-DO - you can make it pick nearly teh same inputs if
    if you add the actual values which are not always the expected values
    inputs = inputs
    responses = responses"""
    whole_inputs = inputs
    averaged_responses = pd.DataFrame()
    averaged_inputs = pd.DataFrame()

    duplicates = [x for x in whole_inputs[whole_inputs.duplicated()].index]
    duplicates_for_averaging = {}
    non_duplicate_list = [x for x in whole_inputs.index if x not in duplicates]
    for non_duplicate in non_duplicate_list:
        this_duplicate_list = []
        non_duplicate_row = whole_inputs.loc[[non_duplicate]]
        for duplicate in duplicates:
            duplicate_row = whole_inputs.loc[[duplicate]]
            if non_duplicate_row.equals(duplicate_row):
                this_duplicate_list.append(duplicate)
                if verbose:
                    print(
                        f"found duplicate pairs: {non_duplicate}, {duplicate}"
                    )
        if len(this_duplicate_list) > 0:
            duplicates_for_averaging[non_duplicate] = this_duplicate_list
        else:
            averaged_inputs = pd.concat([averaged_inputs, non_duplicate_row])
            averaged_responses = pd.concat(
                [averaged_responses, responses.iloc[[non_duplicate]]]
            )

    for non_duplicate, duplicates in duplicates_for_averaging.items():
        # print(f"nd: {non_duplicate}")
        to_average = whole_inputs.loc[[non_duplicate]]
        to_average_responses = responses.loc[[non_duplicate]]
        for duplicate in duplicates:
            to_average = pd.concat([to_average, whole_inputs.loc[[duplicate]]])
            to_average_responses = pd.concat(
                [to_average_responses, responses.loc[[duplicate]]]
            )
        meaned = to_average.mean(axis=0)
        meaned_responses = to_average_responses.mean(axis=0)
        try:
            averaged_inputs = pd.concat(
                [averaged_inputs, pd.DataFrame(meaned).transpose()],
                ignore_index=True,
            )
            averaged_responses = pd.concat(
                [
                    averaged_responses,
                    pd.DataFrame(meaned_responses).transpose(),
                ],
                ignore_index=True,
            )
        except TypeError:
            averaged_inputs = pd.DataFrame(meaned).transpose()
            averaged_responses = pd.DataFrame(meaned_responses).transpose()

    return averaged_inputs, averaged_responses


def calc_ave_coeffs_and_errors(coeffs, labels, errors="std", normalise=False):
    """Coefficient plot
    set error to 'std' for standard deviation
    set error to 'p95' for 95th percentile (
    approximated by 2*std)"""

    ave_coeffs = np.mean(coeffs, axis=0)[0]
    stds = np.std(coeffs, axis=0)[0]
    if normalise:
        ave_coeffs = ave_coeffs / stds
        stds = np.ones_like(ave_coeffs)
    if errors == "std":
        error_bars = stds
    elif errors == "p95":
        # this is an approximation assuming a gaussian distribution in your coeffs
        error_bars = 2 * stds
    else:
        raise ValueError(
            f"Error: errors setting {errors} not known, chose std or p95"
        )

    return ave_coeffs, error_bars


def autotune_model(
    inputs,
    responses,
    source_list,
    response_selector=[0],
    use_scaled_inputs=True,
    do_scaling_here=True,
    drop_duplicates="average",
    errors="p95",
    normalise=True,
    do_hierarchical=True,
    remove_significant=False,
    verbose=False,
):
    """
    inputs: the input matrix
    responses: the results
    response_selector=[0]: which column of results to use, or all of it
    use_scaled_inputs=True: scale model to remove columns baseed on stds
    do_scaling_here=True: if you want scaled inputs and haven't input them
    errors='p95': 95th percentile or 'std' for standard deviation
    normalise=True: setting for coefficient calculation - wants to match scaled inputs
    remove_significant: model will continue removing terms until only one is left
        verbose=False
    """
    sat_inputs = inputs

    if verbose:
        print(f"Source list is {source_list}")
    input_selector = [i for i in range(len(sat_inputs.columns))]
    output_indices = input_selector
    # global list of all column names.
    input_terms = list(sat_inputs.columns)
    output_terms = input_terms
    if verbose:
        print("numbers\tnames")
        for i, v in enumerate(input_terms):
            print(f"{i}\t{v}")
    this_model = None
    have_removed = True
    R2_over_opt = []
    Q2_over_opt = []
    n_terms_over_opt = []
    terms = []

    while have_removed:
        print("Beginning loop")
        selected_input_indices = output_indices
        selected_input_terms = output_terms
        if len(selected_input_indices) == 0:
            break
        data = ModifiableDataSet(sat_inputs, responses)
        if selected_input_terms:
            data.filter(selected_input_terms)
        this_model = AveragedModel(
            data, scale_run_data=True, drop_duplicates=drop_duplicates
        )

        selected_inputs = sat_inputs.iloc[:, selected_input_indices]
        selected_input_terms = list(selected_inputs.columns)
        R2_over_opt.append(this_model.r2)
        Q2_over_opt.append(this_model.q2)

        # cell 2
        # print("Cell 2:")
        print(f"Selected terms {selected_input_terms}")
        print(f"Source List: {source_list}")
        # build a dictionary mapping from input term to the
        # set of derived term's indices.
        # Note that we are only caring about indices still in
        # selected_input_indices (I.e. ones we have not thrown out!)
        dependency_dict = {}
        for i in selected_input_indices:
            dependency_dict[i] = set()
            # ignore 1st order terms. They have no antecedents
            if isinstance(source_list[i], str):
                continue
            # single term - a direct power of a 1st order term.
            if isinstance(source_list[i], int):
                if i in selected_input_indices:
                    dependency_dict[source_list[i]].add(i)
            # some other 2nd+ term.
            if isinstance(source_list[i], list):
                for x in source_list[i]:
                    if i in selected_input_indices:
                        try:
                            dependency_dict[x].add(i)
                        except:  # TODO:: fix blank except. I _think_ KeyError
                            if do_hierarchical:
                                print(
                                    "Error: Heirarchical model missing lower level terms!!!!"
                                )
        print(dependency_dict)
        # Handy shortcut - since the empty set is considered false,
        # we can just test dependency_dict[some_term] to see if there
        # are still dependents.

        # cell 3
        # enforce_hierarchical_model = do_hierarchical
        # whether to enforce hierarchy over terms (i.e. a lower order term
        # must be present for a higher order one)

        ave_coeffs, error_bars = calc_ave_coeffs_and_errors(
            coeffs=this_model.coeffs,
            labels=selected_input_terms,
            errors="p95",
            normalise=True,
        )
        n_terms_over_opt.append(len(ave_coeffs))

        # create a copy of source_list that we will modify for the next iteration
        output_indices = list(selected_input_indices)

        # build a list of all terms whose error bar crosses 0.
        # Values are tuples of the form (source_list index, |error_bar_size|)
        insignificant_terms = []
        for i in range(len(ave_coeffs)):
            if abs(ave_coeffs[i]) < abs(error_bars[i]):
                # print("{:.2}- {:.2}", ave_coeffs[i], error_bars[i])
                if verbose:
                    print(f"{i}:\t{input_terms[i]}\t{source_list[i]}")
                insignificant_terms.append(
                    (selected_input_indices[i], abs(ave_coeffs[i]))
                )  # diffs.append(abs(ave_coeffs[i]) - abs(error_bars[i]))

        # for removing smallest significant terms
        if insignificant_terms == [] and remove_significant:
            for i in range(len(ave_coeffs)):
                # print("{:.2}- {:.2}", ave_coeffs[i], error_bars[i])
                print(f"{i}:\t{input_terms[i]}\t{source_list[i]}")
                insignificant_terms.append(
                    (selected_input_indices[i], abs(ave_coeffs[i]))
                )  # diffs.append(abs(ave_coeffs[i]) - abs(error_bars[i]))

        # Now sort in order of ave_coeff
        insignificant_terms = sorted(insignificant_terms, key=lambda x: x[1])
        print(f"Insignificant Terms: {insignificant_terms}")

        # Now find the first term we can remove (if any)
        have_removed = False

        for idx, error_value in insignificant_terms:
            # If it has dependents, and you're doing an heirarchical model skip it
            if do_hierarchical:
                if dependency_dict[idx]:
                    continue
            print(
                f"removing term {input_terms[idx]} ({idx}) with error {error_value}"
            )
            output_indices.remove(idx)
            output_terms.remove(output_terms[idx])
            have_removed = True
            break

        print(f"output_indices are {output_indices}")
        terms.append(output_indices)
    return (
        output_indices,
        this_model,
    )


def map_chemical_space(
    unscaled_model,
    x_key,
    y_key,
    c_key,
    x_limits,
    y_limits,
    constant,
    n_points,
    hook_function,
):
    min_x = x_limits[0]
    max_x = x_limits[1]
    min_y = y_limits[0]
    max_y = y_limits[1]

    x = np.linspace(min_x, max_x, n_points)
    y = np.linspace(min_y, max_y, n_points)
    # c = np.linspace(constant, constant, n_points)

    mesh_x, mesh_y = np.meshgrid(x, y)

    z_df = pd.DataFrame()
    z_df[x_key] = mesh_x.reshape(-1)
    z_df[y_key] = mesh_y.reshape(-1)
    z_df[c_key] = constant
    # add any extra terms
    z_df = hook_function(z_df)

    mesh_z = unscaled_model.predict(z_df).reshape(n_points, n_points)

    return mesh_x, mesh_y, mesh_z


def map_chemical_space_new(
    unscaled_model,
    x_key,
    y_key,
    c_key,
    x_limits,
    y_limits,
    constant,
    n_points,
    hook_function,
    model,
    inputs,
    input_selector,
    source_list=None,
):
    if source_list is None:
        source_list = []
    min_x = x_limits[0]
    max_x = x_limits[1]
    min_y = y_limits[0]
    max_y = y_limits[1]

    x = np.linspace(min_x, max_x, n_points)
    y = np.linspace(min_y, max_y, n_points)
    # c = np.linspace(constant, constant, n_points)

    mesh_x, mesh_y = np.meshgrid(x, y)
    z_df = pd.DataFrame()
    z_df[x_key] = mesh_x.reshape(-1)
    z_df[y_key] = mesh_y.reshape(-1)
    z_df[c_key] = constant
    z_df = hook_function(z_df)

    mesh_z = unscaled_model.predict(z_df).reshape(n_points, n_points)

    return mesh_x, mesh_y, mesh_z


def add_higher_order_terms(
    inputs: pd.DataFrame,
    add_squares: bool = True,
    add_interactions: bool = True,
    column_list: list = [],
    verbose: bool = True,
):
    """Adds in squares and interactions terms
    inputs: the input/feature/variable array with data
    add_squares=True : whether to add square terms, e.g. x_1^2, x_2^2
    add_interactions=True: whether to add interaction terms, x_1*x_2, etc
    column_list=[]: to select only a subset of columns, input a column list here
    Currently does not go above power of 2

    returns saturated array and a list of which inputs created which column"""

    sat_inputs = copy.deepcopy(inputs)
    if not column_list:
        # do all columns
        column_list = [x for x in inputs.columns]
    if verbose:
        print(f"Input array has columns {column_list}")

    source_list = [x for x in column_list]

    if add_squares:
        if verbose:
            print("Adding square terms:")
        for i in range(len(column_list)):
            source_list.append(i)
            input_name = column_list[i]
            new_name = input_name + "**2"
            if verbose:
                print(new_name)
            sat_inputs[new_name] = inputs[input_name] * inputs[input_name]

    if add_interactions:
        if verbose:
            print("Adding interaction terms:")
        for i in range(len(column_list)):
            for j in range(i + 1, len(column_list)):
                source_list.append([i, j])
                input_name_1 = column_list[i]
                input_name_2 = column_list[j]
                new_name = input_name_1 + "*" + input_name_2
                if verbose:
                    print(new_name)
                sat_inputs[new_name] = (
                    inputs[input_name_1] * inputs[input_name_2]
                )

    return sat_inputs, source_list


def predict_from_model(model, inputs, input_selector):
    """Reorgs the inputs and does a prediction
    model = the model to use
    inputs = the saturated inputs
    input_selector = the subset of inputs the model is using
    """
    list_of_terms = [inputs.columns[x] for x in input_selector]
    model_inputs = inputs[list_of_terms]
    predictions = model.predict(model_inputs)
    return predictions, list_of_terms



from typing import List, Tuple
import pandas as pd
from doenut.data.modifiers.data_set_modifier import DataSetModifier


class ColumnSelector(DataSetModifier):
    """
    DataSet Modifier to remove columns from the dataset
    """

    @classmethod
    def _parse_selector(
        cls, data: pd.DataFrame, selector: List["str | int"]
    ) -> Tuple[List[str], List[int]]:
        """
        Internal helper function to take either a list of column names or
        column indices and convert it to the other.
        @param data: The data set the list applies to
        @param selector: The known selector list
        @return: Tuple of the column names and indices as lists
        """
        if isinstance(selector[0], str):  # columns provided
            # First validate it
            for col in selector:
                if col not in data.columns:
                    raise ValueError(f"Data lacks column {col}")
            return selector, [
                i for i, j in enumerate(data.columns) if j in selector
            ]
        elif isinstance(selector[0], int):  # column indices provided
            for idx in selector:
                if idx >= len(data):
                    raise IndexError(f"Index {idx} out of range for data")
            return [data.columns[i] for i in selector], selector

        raise ValueError("Type of selector needs to be string or int")

    def __init__(
        self,
        inputs: pd.DataFrame,
        responses: pd.DataFrame,
        input_selector: List["str | int"] = None,
        response_selector: List["str | int"] = None,
    ):
        """
        Used to filter specific columns from a dataset. Note that at least one
        of input_selector and response_selector must be specified.
        The selector should be a list of either column names or column indices
        @param inputs: The dataset's inputs
        @param responses: The dataset's responses
        @param input_selector: A list to filter the inputs by
        @param response_selector: A list to filter the responses by

        """
        super().__init__(inputs, responses)
        # Validate inputs
        if input_selector is None and response_selector is None:
            raise ValueError(
                "At least one of input_selector and response_selector is required."
            )

        # Parse / Validate the input selector
        if input_selector is not None:
            (self.input_selector, self.input_indices) = self._parse_selector(
                inputs, input_selector
            )
        else:
            self.input_selector = None
            self.input_indices = None

        if response_selector is not None:
            (
                self.response_selector,
                self.response_indices,
            ) = self._parse_selector(responses, response_selector)
        else:
            self.response_selector = None
            self.response_indices = None

    def apply_to_inputs(self, data: pd.DataFrame) -> pd.DataFrame:
        if not self.input_indices:
            return data
        return data.iloc[:, self.input_indices]

    def apply_to_responses(self, data: pd.DataFrame) -> pd.DataFrame:
        if not self.response_indices:
            return data
        return data.iloc[:, self.response_indices]



"""
doenut.plot

Provides graph plotting functions for DoeNut
"""

from matplotlib import pyplot as plt
import numpy as np
import doenut


def clear_figure():
    """
    Wrapper for matplotlib clear figure to avoid imports elsewhere
    """
    plt.clf()


def replicate_plot(inputs, responses, key):
    """Plots a replicate plot which shows all experiments
    and identifies the replicates
    inputs:
    responses:
    key: column in responses that you wish to plot"""
    plt.title(key)
    replicate_row_list = doenut.find_replicates(inputs)
    non_replicate_row_list = [
        x for x in range(len(responses)) if x not in replicate_row_list
    ]
    col = responses[key]
    x_data = [x for x in range(len(non_replicate_row_list))]
    y_data = [col.loc[x] for x in non_replicate_row_list]
    plt.plot(x_data, y_data, "o")
    ax = plt.gca()
    ax.set_xticks(x_data)
    ax.set_xticklabels(non_replicate_row_list)
    plt.xlabel("Experiment No.")
    plt.ylabel("Response")

    x_data = [
        len(non_replicate_row_list) for _ in range(len(replicate_row_list))
    ]
    y_data = [col.loc[x] for x in replicate_row_list]
    plt.plot(x_data, y_data, "or")
    return


def plot_observed_vs_predicted(
    responses, predictions, range_x=None, label="", do_axes_equal=True
):
    """plots a graph duh
    range should be in the form [min_x, max_x]
    else it will take from responses"""
    if range_x is None:
        range_x = []
    plt.plot(responses, predictions, "o")
    min_xy = np.min([np.min(responses), np.min(predictions)])
    max_xy = np.max([np.max(responses), np.max(predictions)])
    if not range_x:
        range_x = [(min_xy // 10) * 10, (max_xy // 10) * 10 + 10]
    plt.plot(range_x, range_x)
    x_label = "Measured " + label
    y_label = "Predicted " + label
    plt.xlabel(x_label)
    plt.ylabel(y_label)
    if do_axes_equal:
        plt.gca().set_aspect("equal")
    return


def plot_summary_of_fit_small(R2, Q2):
    """Plots a nice graph of R2 and Q2"""
    my_colors = [
        "green",
        "blue",
        "pink",
        "cyan",
        "red",
        "green",
        "blue",
        "cyan",
        "orange",
        "purple",
    ]
    plt.bar([1, 2], [R2, Q2], color=my_colors)
    plt.grid(True)
    plt.yticks([0, 0.2, 0.4, 0.6, 0.8, 1.0])
    # plt.ylabel('Average Price')
    plt.xticks([1, 2], labels=["R2", "Q2"])
    plt.ylim((0, 1))
    plt.grid(axis="x")
    return


def coeff_plot(coeffs, labels, errors="std", normalise=False):
    """Coefficient plot
    set error to 'std' for standard deviation
    set error to 'p95' for 95th percentile (
    approximated by 2*std)"""
    # get values
    ave_coeffs, error_bars = doenut.calc_ave_coeffs_and_errors(
        coeffs, labels, errors=errors, normalise=normalise
    )
    # plot values
    x_points = [x for x in range(len(ave_coeffs))]
    f = plt.figure()
    f.set_figwidth(16)
    f.set_figheight(9)
    print(f"Input_selector was: {x_points}")
    print(f"Average coefficients are: {ave_coeffs}")
    print(f"Errors are {error_bars}")
    # print(errors)
    print(f"Coefficient labels are: {labels}")
    plt.bar(x_points, ave_coeffs, yerr=error_bars, capsize=20)
    plt.xticks(x_points, labels)
    return


def four_D_contour_plot(
    unscaled_model,
    x_key,
    y_key,
    c_key,
    x_limits,
    y_limits,
    constants,
    n_points,
    my_function,
    fig_label="",
    input_selector=None,
    x_label="",
    y_label="",
    constant_label="",
    z_label="",
    cmap="jet",
    num_of_z_levels=9,
    z_limits=None,
    tidy_subfig_axes=False,
):
    """This could be improved to take any number of data
    1. unscaled_model: the model you just trained
    2. x_key: name in the dataframe for the input to go on the x-axis
    3. y_key: name in the dataframe for the input to go on the y-axis
    4. c_key: name in the dataframe for the input to be the constant for each plot (i.e. equivalents of pyrollidine)
    5. x_limits: limits of the x-axis: min and max time values
    6. y_limits: limits of the y-axis: min and max temperatures
    7. constants: values of pyrollidine to keep constant for the 3 plots
    8. n_points: how many points in the x and y direction to use to build the map
    9. my_function: a little function to add higher order terms if the model requires it
    10. fig_label: label for the overall figure
    11. x_label: label for x-axes
    12. y_label: label for y-axis
    13. constant_label: label for top of subplots
    14: z_label: label for the heatbar
    15: cmap: colourmap for the plot (yes you can change it, do not spend hours playing around with the colourscheme!)
    16: num_of_z_levels: number of levels for the contours. You will want one more than you think you do
    17: z_limits: limits for the yield, i.e. minimum and maximum."""

    if input_selector is None:
        input_selector = []
    if z_limits is None:
        z_limits = []
    X_1, Y_1, Z_1 = doenut.map_chemical_space(
        unscaled_model=unscaled_model,
        x_key=x_key,
        y_key=y_key,
        c_key=c_key,
        x_limits=x_limits,
        y_limits=y_limits,
        constant=constants[0],
        n_points=n_points,
        hook_function=my_function,
    )
    X_2, Y_2, Z_2 = doenut.map_chemical_space(
        unscaled_model=unscaled_model,
        x_key=x_key,
        y_key=y_key,
        c_key=c_key,
        x_limits=x_limits,
        y_limits=y_limits,
        constant=constants[1],
        n_points=n_points,
        hook_function=my_function,
    )
    X_3, Y_3, Z_3 = doenut.map_chemical_space(
        unscaled_model=unscaled_model,
        x_key=x_key,
        y_key=y_key,
        c_key=c_key,
        x_limits=x_limits,
        y_limits=y_limits,
        constant=constants[2],
        n_points=n_points,
        hook_function=my_function,
    )

    if x_label == "":
        x_label = x_key
    if y_label == "":
        y_label = y_key
    if z_label == "":
        z_label = fig_label
    if constant_label == "":
        constant_label = c_key

    if not z_limits:
        z_min = np.min([np.min(Z_1), np.min(Z_2), np.min(Z_3)])
        z_max = np.max([np.max(Z_1), np.max(Z_2), np.max(Z_3)])
    else:
        z_min = z_limits[0]
        z_max = z_limits[1]

    plt.figure(figsize=(20, 12))
    num_of_levels = 9
    levels = np.linspace(z_min, z_max, num_of_z_levels)
    # levels = np.linspace(20, 100, num_of_levels)
    fig, (ax1, ax2, ax3) = plt.subplots(1, 3)
    fig.suptitle(fig_label)
    egg = ax1.contourf(X_1, Y_1, Z_1, num_of_levels, levels=levels, cmap=cmap)

    egg1 = ax1.contour(
        X_1, Y_1, Z_1, num_of_levels, levels=levels, colors="black"
    )
    if np.max(levels) > 10:
        plt.clabel(egg1, fontsize=16, inline=1, fmt="%1.0f")
    else:
        plt.clabel(egg1, fontsize=16, inline=1, fmt="%1.2f")
    # egg.xlabel('egg')
    ax2.contourf(X_2, Y_2, Z_2, num_of_levels, levels=levels, cmap=cmap)
    egg2 = ax2.contour(
        X_2, Y_2, Z_2, num_of_levels, levels=levels, colors="black"
    )
    if np.max(levels) > 10:
        plt.clabel(egg2, fontsize=16, inline=1, fmt="%1.0f")
    else:
        plt.clabel(egg2, fontsize=16, inline=1, fmt="%1.2f")
    ax3.contourf(X_3, Y_3, Z_3, num_of_levels, levels=levels, cmap=cmap)
    egg3 = ax3.contour(
        X_3, Y_3, Z_3, num_of_levels, levels=levels, colors="black"
    )
    if np.max(levels) > 10:
        plt.clabel(egg3, fontsize=16, inline=1, fmt="%1.0f")
    else:
        plt.clabel(egg3, fontsize=16, inline=1, fmt="%1.2f")
    # make constant label for subplot title
    ax1.set_title(constant_label + " = " + str(constants[0]))
    ax2.set_title(constant_label + " = " + str(constants[1]))
    ax3.set_title(constant_label + " = " + str(constants[2]))
    # make constant label for subplot title
    ax1.set_xlabel(x_label)
    ax2.set_xlabel(x_label)
    ax3.set_xlabel(x_label)

    ax1.set_ylabel(y_label)

    if tidy_subfig_axes:
        ax2.yaxis.set_visible(False)
        ax3.yaxis.set_visible(False)

        # fig.colorbar()  # ax1.clabel(contours, inline=True, fontsize=12)

    fig.subplots_adjust(right=0.9)
    cbar_ax = fig.add_axes([0.95, 0.15, 0.05, 0.7])
    cbar = fig.colorbar(egg, cax=cbar_ax)
    # cbar.set_label(z_label, rotation=270)
    cbar.set_label(z_label, labelpad=0, y=1.10, rotation=0)

    return


def plot_training(R2_over_opt, Q2_over_opt, n_terms_over_opt):
    """Plots optimisation correlation coefficient outcomes
    n_terms_over_opt
    R2_over_opt: list of R2 over optimisation
    Q2_over_opt: list of Q2 over optimisation
    n_terms_over_opt: running number of terms

    """
    ax = plt.axes()
    x_data = range(len(R2_over_opt))
    plt.plot(x_data, R2_over_opt)
    plt.plot(x_data, Q2_over_opt)
    # plt.set_xlabels(n_terms_over_opt)
    ax.set_xticks(x_data)
    if not n_terms_over_opt == []:
        ax.set_xticklabels(n_terms_over_opt)
    plt.legend(["R$^2$", "Q$^2$"])
    plt.ylim([0, 1])
    plt.plot([x_data[0], x_data[-1]], [0.5, 0.5], "-.")
    plt.xlabel("Number of terms")
    plt.ylabel("Correlation coefficient")
    return ax



import numpy as np
import pandas as pd
from typing import Tuple
from doenut.data.modifiers.data_set_modifier import DataSetModifier


class OrthoScaler(DataSetModifier):
    """
    Takes a dataset and scales it per column using an ortho scaling to
    the range -1 ... 1
    By default only inputs are scaled, but this can be overridden.
    """

    @classmethod
    def _compute_scaling(cls, data: pd.DataFrame) -> Tuple[float, float]:
        data_max = np.max(data, axis=0)
        data_min = np.min(data, axis=0)
        mj = (data_min + data_max) / 2
        rj = (data_max - data_min) / 2
        return mj, rj

    def __init__(
        self,
        inputs: pd.DataFrame,
        responses: pd.DataFrame,
        scale_responses: bool = False,
    ) -> None:
        """
        Use this modifier to add a standard ortho scaling to the dataset, so
        it has a range of -1..1
        @param inputs: The inputs of the dataset
        @param responses: The responses of the dataset
        @param scale_responses: Whether to also scale the responses
        """
        super().__init__(inputs, responses)
        self.inputs_mj, self.inputs_rj = self._compute_scaling(inputs)
        self.responses_mj, self.responses_rj = self._compute_scaling(responses)
        self.scale_responses = scale_responses

    def apply_to_inputs(self, data: pd.DataFrame) -> pd.DataFrame:
        return (data - self.inputs_mj) / self.inputs_rj

    def apply_to_responses(self, data: pd.DataFrame) -> pd.DataFrame:
        if not self.scale_responses:
            return data
        return (data - self.responses_mj) / self.responses_rj
